"""

"""

from pyspark.sql import DataFrame
from univariate.analyzer import AnalysisReport
from univariate.missing_value.helper import ParseReportException
from univariate.missing_value.strategy.unmarked_detection import UnmarkedMissingValueDetectionStrategyType, UnmarkedMissingValueDetectionStrategy
from univariate.missing_value import MissingValueReport
from typing import Optional
import pyspark.sql.functions as F
import logging
import importlib

logger = logging.getLogger()


class MissingValueDetector:
    """
    Detect missing value in univariate time series
    If time series is 'regular', it can detect marked and unmarked missing value.


    If time series is 'irregular' for now, it can only detect marked missing value, not unmarked.
    """
    # todo: support 'irregular' time series mssing value processing
    regularity_report: AnalysisReport
    unmarked_strategy: UnmarkedMissingValueDetectionStrategy

    # todo: after refactoring, convert regularity as enum val
    def __init__(self, regularity: str, unmarked_detection_strategy_type: Optional[UnmarkedMissingValueDetectionStrategyType]=None, marked_only: bool=False):
        """

        :param regularity:
        :param unmarked_detection_strategy_type: Enum type. If it is None strategy will depend on default strategy
        :param marked_only: If marked_only is True, then detect both marked and unmarked missing value.
                            if marked_only is False, then detect only marked missing value.
        """
        self.__instantiate_unmarked_detector(regularity, unmarked_detection_strategy_type, marked_only)

    def __instantiate_unmarked_detector(self, regularity: str, strategy_type: Optional[UnmarkedMissingValueDetectionStrategyType], marked_only: bool):
        logger.debug("__instatntiate_unmarked_detector is called")
        if marked_only:
            strategy_type = UnmarkedMissingValueDetectionStrategyType.do_not_detect
        elif strategy_type is None and regularity == "regular":
            strategy_type = UnmarkedMissingValueDetectionStrategyType.divide_and_round  # todo: configurable
        elif strategy_type is None and regularity == "irregular":
            strategy_type = UnmarkedMissingValueDetectionStrategyType.do_not_detect  # todo: configurable
        logger.info(f"unmarked_detection strategy type: {strategy_type.value}")
        print(f"unmarked_detection strategy type: {strategy_type.value}")

        # Instantiate
        concrete_cls = getattr(
            importlib.import_module("univariate.missing_value.strategy.unmarked_detection"),
            strategy_type.value,
        )
        self.unmarked_strategy = concrete_cls()

    def detect_missing_values(self, ts: DataFrame, time_col_name: str, data_col_name: str, period: int) -> MissingValueReport:
        """

        :param ts:
        :param time_col_name:
        :param data_col_name:
        :return:
        """
        # todo: how to handle time col missing value (deletion?)
        # todo: enhance report
        report = MissingValueReport()

        # 1. marked missing values
        report.marked["count"] = ts.select([F.count(F.when(F.isnan(data_col_name) | F.col(data_col_name).isNull() , True))]).first()[0]

        # 2. unmarked missing values
        marked_df = self.unmarked_strategy.detect(ts=ts, time_col_name=time_col_name, data_col_name=data_col_name, period=period).cache()
        report.unmarked["count"] = marked_df.count() - ts.count()
        report.unmarked["marking_df"] = marked_df

        # 3. total
        report.total["count"] = report.marked["count"] + report.unmarked["count"]
        # todo: enhance report
        return report



"""

"""

import logging
from univariate.ts_validator import Validator
from pyspark.sql import DataFrame
import pyspark.sql.functions as F
from pyspark.sql.window import Window
from univariate.ts_validator.helper import NonMonotonicForwardException

logger = logging.getLogger()


class ForwardValidator(Validator):
    """
    Validate ts has monotonic forward timestamps only, not backward and equal
    """

    def validate(self, ts: DataFrame, time_col_name: str):
        logger.debug("Forward validator called")
        lagged_time_ts = (
            ts.select(time_col_name)
            .withColumn("idx", F.monotonically_increasing_id())
            .withColumn(
                "lagged_time",
                F.lag(time_col_name).over(Window.orderBy("idx")),
            )
        )
        non_forward_ts = (
            lagged_time_ts.withColumn(
                "diff",
                lagged_time_ts[time_col_name] - lagged_time_ts["lagged_time"],
            )
            .filter(F.col("diff") <= 0.0)
            .cache()
        )
        if non_forward_ts.count() > 0:  # invalid ts
            msg = (
                "Time series has invalid timestamp, with non-monotonic forward increasing. check this\n"
                + non_forward_ts.toPandas().to_string()
            )
            non_forward_ts.unpersist()
            logger.error(msg)
            raise NonMonotonicForwardException(msg=msg)
        logger.info("Forward validation success!")
        non_forward_ts.unpersist()



"""

"""

from abc import ABCMeta, abstractmethod
from univariate.analyzer import AnalysisReport
from typing import Dict, Optional
from enum import Enum


class Analyzer(metaclass=ABCMeta):
    """
    Abstract analyzer for analyzing with specific strategy.
    """

    @abstractmethod
    def analyze(self) -> AnalysisReport:
        """
            Command Executor
        :return:
        """
        pass



"""

"""

from univariate.missing_value.strategy.handle_missing_value import (
    MissingValueHandleStrategy,
)
from pyspark.sql import DataFrame
import pyspark.sql.functions as F


class MeanImputation(MissingValueHandleStrategy):
    """ """

    def handle(
        self, missed_ts: DataFrame, time_col_name: str, data_col_name
    ) -> DataFrame:
        return missed_ts.fillna(
            value=round(missed_ts.agg(F.mean(data_col_name)).first()[0], 3),
            subset=data_col_name,
        )



from univariate.missing_value.missing_value_report import MissingValueReport
from univariate.missing_value.missing_value_detector import MissingValueDetector
from univariate.missing_value.missing_value_handler import MissingValueHandler



"""

"""

from enum import Enum
from abc import ABCMeta
from pyspark.sql import DataFrame
from univariate.analyzer import AnalysisReport


class UnmarkedMissingValueDetectionStrategyType(Enum):
    divide_and_round = "DivideAndRound"
    do_not_detect = "DoNotDetect"


class UnmarkedMissingValueDetectionStrategy(metaclass=ABCMeta):
    """ """

    @staticmethod
    def detect(
        ts: DataFrame,
        time_col_name: str,
        data_col_name: str,
        period: int
    ) -> DataFrame:
        pass



"""

"""

from enum import Enum
from abc import ABCMeta, abstractmethod
from pyspark.sql import DataFrame


class MissingValueHandleStrategyType(Enum):
    deletion = "Deletion"
    mean_imputation = "MeanImputation"
    median_imputation = "MedianImputation"
    mode_imputation = "ModeImputation"
    zero_imputation = "ZeroImputation"
    constant_imputation = "ConstantImputation"
    # k_nn_imputation = "KNNImputation" #multi
    # mice_imputation = "MICEImputation" #multi
    hot_deck_imputation = "HOT_DECK_IMPUTATION"
    datawig = "Datawig"
    interpolation = "Interpolation"  # todo: SOTA
    # todo: continous update
    # todo: where cross-informated strategy will be
    # todo: enhance
    # todo: process exceptions with sufficient test coverage


class MissingValueHandleStrategy(metaclass=ABCMeta):
    """
    Missing
    """

    @abstractmethod
    def handle(
        self, missed_ts: DataFrame, time_col_name: str, data_col_name
    ) -> DataFrame:
        pass



from univariate.missing_value.strategy.handle_missing_value.missing_value_handle_strategy import (
    MissingValueHandleStrategyType,
)
from univariate.missing_value.strategy.handle_missing_value.missing_value_handle_strategy import (
    MissingValueHandleStrategy,
)
from univariate.missing_value.strategy.handle_missing_value.deletion import Deletion
from univariate.missing_value.strategy.handle_missing_value.mean_imputation import (
    MeanImputation,
)



"""
Standard output hook
"""

from univariate.hook import Hook
from univariate.analyzer import AnalysisReport


class StdoutHook(Hook):
    """
    Stdout Concrete hook for develop
    """

    def do_post_analysis(self, report: AnalysisReport):
        """
        Print metrics, parameters, plots

        :return:
        """
        print("parameters:")
        for parameter_key in report.parameters.keys():
            print(f"/t{parameter_key}: {report.parameters[parameter_key]}")
        print()

        print("metrics:")
        for metric_key in report.metrics.keys():
            print(f"/t{metric_key}: {report.metrics[metric_key]}")
        print()

        print("plot:")
        for plot_key in report.plots.keys():
            report.plots[plot_key].show()



"""

"""

from abc import ABCMeta, abstractmethod
from univariate.analyzer import AnalysisReport


class Hook(metaclass=ABCMeta):
    """
    Hook for some act in post analysis, such as storing result, notifying.
    It is intended for multi-threading in future.
    """

    @abstractmethod
    def do_post_analysis(self, report: AnalysisReport):
        pass



"""
 Descriptive Statistics analyzer
"""

from pyspark.sql import DataFrame
import pyspark.sql.functions as F
from pyspark import StorageLevel
from pyspark.sql.types import TimestampType
from univariate.analyzer import Analyzer, AnalysisReport
import logging
import plotly.express as px
import plotly.figure_factory as ff
from scipy.stats import gaussian_kde


logger = logging.getLogger()


class DescriptiveStatAnalyzer(Analyzer):
    """
    Descriptvie Statistics analyzer
    Items
      - count
      - mean (arithmetic, geometric, harmonic)
      - mode, median
      - max, min
      - std dev
      - skewness
      - kurtosis
      - coefficient of variation
    """

    def __init__(self):
        """ """

    def analyze(self, ts: DataFrame, **kwargs) -> AnalysisReport:
        """

        :return:
        """
        # todo: refactor long method
        logger.debug("analyze called")
        ts_df = (
            ts.select(kwargs["data_col_name"])
            .dropna()
            .persist(StorageLevel.MEMORY_ONLY)
        )
        ts_np_array = ts_df.toPandas().to_numpy()
        report = AnalysisReport()

        # 1. Aggregate (max, min, mean, median, count, stddev, skewness, kurtosis)
        stats_dict = (
            ts_df.agg(
                F.max(kwargs["data_col_name"]).alias("max"),
                F.min(kwargs["data_col_name"]).alias("min"),
                F.count(kwargs["data_col_name"]).alias("count"),
                F.mean(kwargs["data_col_name"]).alias("mean"),
                F.percentile_approx(kwargs["data_col_name"], percentage=0.5).alias(
                    "median"
                ),
                F.percentile_approx(kwargs["data_col_name"], percentage=0.25).alias(
                    "Q1"
                ),
                F.percentile_approx(kwargs["data_col_name"], percentage=0.7).alias(
                    "Q3"
                ),
                F.stddev_samp(kwargs["data_col_name"]).alias("stddev"),
                F.skewness(kwargs["data_col_name"]).alias("skewness"),
                F.kurtosis(kwargs["data_col_name"]).alias("kurtosis"),
            )
            .first()
            .asDict()
        )

        # 2. mode # todo: except iid
        stats_dict["mode"] = (
            ts_df.groupBy(kwargs["data_col_name"])
            .count()
            .orderBy("count", ascending=False)
            .first()[0]
        )

        # 3. coefficient of variation
        stats_dict["cv"] = stats_dict["stddev"] / stats_dict["mean"]

        report.parameters = stats_dict.copy()

        # 4. histogram (with max, min, mean, median, mode)
        histogram = px.histogram(ts_df.toPandas(), x="input_data", marginal="violin")
        histogram.add_vline(
            x=stats_dict["min"], annotation_text="min", line_width=1, line_dash="dash"
        )
        histogram.add_vline(
            x=stats_dict["max"], annotation_text="max", line_width=1, line_dash="dash"
        )
        histogram.add_vline(
            x=stats_dict["mode"], annotation_text="mode", line_width=1, line_dash="dash"
        )
        report.plots["histogram"] = histogram

        # 2. timeseries plot with range slider
        timeseries_plot = px.line(
            ts.withColumn(
                kwargs["time_col_name"],
                (ts[kwargs["time_col_name"]] / 1000).cast(TimestampType()),
            )
            .select(kwargs["data_col_name"], kwargs["time_col_name"])
            .toPandas(),
            x=kwargs["time_col_name"],
            y=kwargs["data_col_name"],
            title="Time Series with Range Slider and Selector",
        )
        timeseries_plot.update_xaxes(
            rangeslider_visible=True,
            rangeselector=dict(
                buttons=list(
                    [
                        dict(count=1, label="1m", step="month", stepmode="backward"),
                        dict(count=6, label="6m", step="month", stepmode="backward"),
                        dict(count=1, label="YTD", step="year", stepmode="todate"),
                        dict(count=1, label="1y", step="year", stepmode="backward"),
                        dict(step="all"),
                    ]
                )
            ),
        )
        report.plots["timeseries_plot"] = timeseries_plot

        # 5. dist plot (density curve, skewness, median, mean, mode)
        pallete = px.colors.qualitative.Plotly
        dist_plot = ff.create_distplot(
            ts_np_array.reshape((1, -1)), group_labels=["input_data"], show_hist=False
        )  # from plotly-python v5.10.0, curve is estimated 'kde'
        dist_plot.update_layout(title_text="Density Curve")
        kde = gaussian_kde(ts_np_array.reshape((-1,)))
        dist_plot.add_shape(
            type="line",
            x0=stats_dict["mean"],
            x1=stats_dict["mean"],
            y0=0.0,
            y1=kde(stats_dict["mean"])[0],
            opacity=0.5,
            name="mean",
            line_dash="dash",
            line=dict(color=pallete[0], width=2),
        )
        dist_plot.add_annotation(
            x=stats_dict["mean"],
            y=kde(stats_dict["mean"])[0],
            text="mean",
            font=dict(color=pallete[0]),
            arrowcolor=pallete[0],
        )
        dist_plot.add_shape(
            type="line",
            x0=stats_dict["mean"],
            x1=stats_dict["mean"],
            y0=0.0,
            y1=kde(stats_dict["mean"])[0],
            opacity=0.5,
            name="mode",
            line_dash="dash",
            line=dict(color=pallete[1], width=2),
        )
        dist_plot.add_annotation(
            x=stats_dict["mean"],
            y=kde(stats_dict["mean"])[0],
            text="mode",
            font=dict(color=pallete[1]),
            arrowcolor=pallete[1],
        )
        dist_plot.add_shape(
            type="line",
            x0=stats_dict["median"],
            x1=stats_dict["median"],
            y0=0.0,
            y1=kde(stats_dict["median"])[0],
            opacity=0.5,
            name="median",
            line_dash="dash",
            line=dict(color=pallete[2], width=2),
        )
        dist_plot.add_annotation(
            x=stats_dict["median"],
            y=kde(stats_dict["median"])[0],
            text="median",
            font=dict(color=pallete[2]),
            arrowcolor=pallete[2],
        )
        dist_plot.add_annotation(
            text=f"<b>mean</b>: {round(stats_dict['mean'], 3)}<br>"
            f"<b>mode</b>: {round(stats_dict['mode'], 3)}<br>"
            f"<b>Q1</b>: {round(stats_dict['Q1'], 3)}<br>"
            f"<b>Q2(median)</b>: {round(stats_dict['median'], 3)}<br>"
            f"<b>Q3</b>: {round(stats_dict['Q3'], 3)}<br>"
            f"<b>skewness</b>: {round(stats_dict['skewness'], 3)}<br>"
            f"<b>kurtosis</b>: {round(stats_dict['kurtosis'], 3)}<br>"
            f"<b>coefficient of variation</b>: {round(stats_dict['cv'], 3)}",
            align="right",
            showarrow=False,
            xref="paper",
            x=1,
            yref="paper",
            y=1,
        )
        report.plots["dist_plot"] = dist_plot

        return report



"""

"""

from univariate.missing_value.strategy.handle_missing_value import (
    MissingValueHandleStrategy,
)
from pyspark.sql import DataFrame


class Deletion(MissingValueHandleStrategy):
    """
    Delete data missing values in dataframe(na, null)
    It does not drop time col missing values
    """

    def handle(
        self, missed_ts: DataFrame, time_col_name: str, data_col_name
    ) -> DataFrame:
        return missed_ts.dropna(subset=data_col_name)



"""

"""

from pyspark.sql import DataFrame
from univariate.missing_value import MissingValueReport
from univariate.missing_value.strategy.handle_missing_value import MissingValueHandleStrategy
import importlib

class MissingValueHandler:
    """
    It can handle missing value with some strategies, such as imputation, deletion, filling constant, interpolation, and so on
    """
    def handle_missing_value(self, missed_ts: DataFrame, time_col_name: str, data_col_name: str) -> DataFrame:
        """

        :param ts:
        :param report:
        :return:
        """
        # todo: decision logic decoupling, decision rule-engine?
        # Instantiate
        concrete_cls = getattr(
            importlib.import_module("univariate.missing_value.strategy.handle_missing_value"),
            "MeanImputation"
        )
        handling_strategy: MissingValueHandleStrategy = concrete_cls()
        return handling_strategy.handle(missed_ts=missed_ts, time_col_name=time_col_name, data_col_name=data_col_name)



"""

"""

from univariate.missing_value.strategy.handle_missing_value import (
    MissingValueHandleStrategy,
)
from pyspark.sql import DataFrame


class Interpolation(MissingValueHandleStrategy):
    """
    # todo : specify interpolation methods, and think about how to deal with sporadic data?
    https://medium.com/delaware-pro/interpolate-big-data-time-series-in-native-pyspark-d270d4b592a1
    """

    # todo: if ts is sporadic, then strategy decision must block use this

    def handle(
        self, missed_ts: DataFrame, time_col_name: str, data_col_name
    ) -> DataFrame:
        pass



"""

"""

from univariate.ts_validator import Validator
from pyspark.sql import DataFrame
from univariate.analyzer import AnalysisReport


class MockValidator(Validator):
    """
    Mock Validator
    """

    def validate(self, ts: DataFrame, time_col_name: str) -> AnalysisReport:
        """

        :param ts:
        :param time_col_name:
        :return:
        """
        report = AnalysisReport()
        report.parameters["validated_df"] = ts
        return report



"""

"""

from univariate.analyzer import Analyzer
from univariate.analyzer import AnalysisReport
from pyspark.sql import DataFrame
import plotly.express as px
from enum import Enum, auto
from statsmodels.tsa.seasonal import STL


class SeasonType(Enum):
    HOURLY = auto()
    DAILY = auto()
    WEEKLY = auto()
    MONTHLY = auto()
    QUARTERLY = auto()
    YEARLY = auto()

class DecompositionAnalyzer(Analyzer):
    """
    Decompose Time series
    """
    def __init__(self):
        """

        """
        # todo: decide decomposition strategy
        pass

    def analyze(self, ts: DataFrame, time_col_name: str, data_col_name: str, ) -> AnalysisReport:
        """

        :param ts:
        :param time_col_name:
        :param data_col_name:
        :return:
        """
        period = self.__calc_period(ts, time_col_name)
        stl = STL(ts.select(data_col_name).toPandas().values.reshape((-1,)), period=period).fit()  # todo: strategies for ts decompositions  #todo: multi seasonal

        report = AnalysisReport()
        report.parameters["observed"] = stl.observed
        report.parameters["trend"] = stl.trend
        report.parameters["seasonal"] = stl.seasonal
        report.parameters["resid"] = stl.resid

        report.plots["observed"] = px.line(stl.observed)
        report.plots["seasonal"] = px.line(stl.seasonal)
        report.plots["trend"] = px.line(stl.trend)
        report.plots["resid"] = px.line(stl.resid)

        return report

    def __calc_period(self, ts: DataFrame, time_col_name: str, season_type: SeasonType=SeasonType.WEEKLY):
        # todo:
        temp = 144
        return temp
